#!/usr/bin/env python3

SEP = ","


def ports2list(ports):
    if ports is None:
        l = []
    else:
        try:
            l = [int(x) for x in ports.split(SEP)]
        except ValueError:
            l = []
    return l


def ports2byte(ports):
    out = 0
    l = ports2list(ports)
    if l == []:
        out = 0
    else:
        for i in l:
            out |= 1 << (int(i) - 1)
    return out


def byte2ports(byte):
    out = []
    for i in range(32):
        if byte % 2:
            out.append(str(i + 1))
        byte >>= 1
    return SEP.join(out)


def mac_to_bytes(mac):
    return bytes(int(byte, 16) for byte in mac.split(":"))


def mac_to_str(mac):
    return ":".join(format(s, "02x") for s in mac)


if __name__ == "__main__":
    a = ports2byte("1,2,5,6,8,12,15")
    print(a, byte2ports(a))
    print(ports2list("1,2"))
    m = mac_to_bytes("ba:ff:ee:ff:ac:ee")
    print(mac_to_str(m))



#!/usr/bin/env python

import random
import logging
import argparse

from protocol import Protocol
from network import Network, ConnectionProblem, InterfaceProblem
from loglevel import loglevel

logger = logging.getLogger(__name__)


def discover_switches(interface=None):
    net = Network(interface)
    net.send(Protocol.DISCOVERY, {})
    ret = []
    while True:
        try:
            header, payload = net.receive()
            ret.append((net.ip_address, net.host_mac, header, payload))
        except ConnectionProblem:
            break
    return ret


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--interface", "-i")
    parser.add_argument("--command", "-c", action="store_true")
    parser.add_argument("--loglevel", "-l", type=loglevel, default="INFO")
    args = parser.parse_args()
    logging.basicConfig(level=args.loglevel)
    try:
        switches = discover_switches(args.interface)
    except InterfaceProblem as e:
        print("Error:", e)
    else:
        for ip, mac, header, payload in switches:
            if args.command:
                p = {x[1]: x[2] for x in payload}
                cmd = f"./smrt.py --username admin --password admin --host-mac={mac} --ip-address={ip} --switch-mac {p['mac']}"
                print(cmd)
            else:
                print(ip, mac, *payload, sep="\n")
                print("-" * 16)


if __name__ == "__main__":
    main()



"""Provide network interfacing functions."""

import netifaces
import socket, random, logging

from protocol import Protocol
from binary import byte2ports, mac_to_str, mac_to_bytes

logger = logging.getLogger(__name__)


class ConnectionProblem(Exception):
    pass


class InterfaceProblem(Exception):
    pass


class Network:

    BROADCAST_ADDR = "255.255.255.255"
    UDP_SEND_TO_PORT = 29808
    UDP_RECEIVE_FROM_PORT = 29809

    def __init__(self, interface=None, switch_mac="00:00:00:00:00:00"):

        # Normally, this module will be initialized with the MAC address of the switch we want to talk to
        # There are however two other modes that might be used:
        # - Specify MAC address fe:ff:ff:ff:ff:ff to go into broadcast listen mode. We use this to snoop on bidirectional traffic
        # - Specify MAC address ff:ff:ff:ff:ff:ff to go into "fake switch" mode, where we reply to other clients

        self.switch_mac = switch_mac
        self.ip_address, self.host_mac = self.get_interface(interface)

        self.sequence_id = random.randint(0, 1000)

        self.header = Protocol.header["blank"].copy()
        self.header.update(
            {
                "sequence_id": self.sequence_id,
                "host_mac": mac_to_bytes(self.host_mac),
                "switch_mac": mac_to_bytes(self.switch_mac),
            }
        )

        # Sending socket
        self.ss = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        if switch_mac == "fe:ff:ff:ff:ff:ff" or switch_mac == "ff:ff:ff:ff:ff:ff":
            self.ss.bind((Network.BROADCAST_ADDR, Network.UDP_SEND_TO_PORT))
            self.ss.settimeout(10)
        else:
            self.ss.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            self.ss.bind((self.ip_address, Network.UDP_RECEIVE_FROM_PORT))

        # Receiving socket
        self.rs = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        if switch_mac == "ff:ff:ff:ff:ff:ff":
            # This is a signal that we'll be operating in fake switch mode, rather than sending out commands
            # For this, we need to switch the way that the recieving socket binds, to bind locally instead.
            self.rs.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            self.rs.bind((self.ip_address, Network.UDP_SEND_TO_PORT))

        else:

            # Receiving socket
            self.rs.bind((Network.BROADCAST_ADDR, Network.UDP_RECEIVE_FROM_PORT))
            self.rs.settimeout(10)

    def get_interface(self, interface=None) -> list | tuple:
        if interface is None:
            interfaces = netifaces.interfaces()
            if "lo" in interfaces:
                interfaces.remove("lo")
            if len(interfaces) > 1:
                return interfaces
            return []

        settings = []
        addrs = netifaces.ifaddresses(interface)
        logger.debug("addrs:" + repr(addrs))
        if netifaces.AF_INET not in addrs:
            raise InterfaceProblem("not AF_INET address")
        if netifaces.AF_LINK not in addrs:
            raise InterfaceProblem("not AF_LINK address")

        mac = addrs[netifaces.AF_LINK][0]["addr"]
        # take first address of interface
        addr = addrs[netifaces.AF_INET][0]
        if "broadcast" not in addr or "addr" not in addr:
            raise InterfaceProblem("no addr or broadcast for address")
        ip = addr["addr"]

        logger.debug("get_interface: %s %s %s " % (interface, ip, mac))

        return ip, mac

    def send(self, op_code, payload):
        self.sequence_id = (self.sequence_id + 1) % 1000
        self.header.update(
            {
                "sequence_id": self.sequence_id,
                "op_code": op_code,
            }
        )
        packet = Protocol.assemble_packet(self.header, payload)
        logger.debug("Sending Packet: " + packet.hex())
        packet = Protocol.encode(packet)
        logger.debug("Sending Header:  " + str(self.header))
        logger.debug("Sending Payload: " + str(payload))
        if self.switch_mac == "ff:ff:ff:ff:ff:ff":
            self.rs.sendto(
                packet, (Network.BROADCAST_ADDR, Network.UDP_RECEIVE_FROM_PORT)
            )
        else:
            self.ss.sendto(packet, (Network.BROADCAST_ADDR, Network.UDP_SEND_TO_PORT))

    def setHeader(self, header):
        self.header = header

    def receive(self):
        data = self.receive_socket(self.rs)
        if data:
            data = Protocol.decode(data)
            logger.debug("Receive Packet: " + data.hex())
            header, payload = Protocol.split(data)
            header, payload = Protocol.interpret_header(
                header
            ), Protocol.interpret_payload(payload)
            logger.debug("Received Header:  " + str(header))
            logger.debug("Received Payload: " + str(payload))
            self.header["token_id"] = header["token_id"]
            return header, payload
        else:
            raise ConnectionProblem()

    def receive_socket(self, socket):
        data = False
        try:
            data, addr = socket.recvfrom(1500)
        except:
            return False
        return data

    def query(self, op_code, payload):
        self.send(op_code, payload)
        header, payload = self.receive()
        return header, payload

    def login_dict(self, username, password):
        return [
            (Protocol.get_id("username"), username.encode("ascii") + b"\x00"),
            (Protocol.get_id("password"), password.encode("ascii") + b"\x00"),
        ]

    def login(self, username, password):
        self.query(Protocol.GET, [(Protocol.get_id("get_token_id"), b"")])
        self.query(Protocol.LOGIN, self.login_dict(username, password))

    def set(self, username, password, payload):
        self.query(Protocol.GET, [(Protocol.get_id("get_token_id"), b"")])
        real_payload = self.login_dict(username, password)
        real_payload += payload
        header, payload = self.query(Protocol.LOGIN, real_payload)
        return header, payload



#!/usr/bin/env python

import socket, time, random, argparse, logging

from protocol import Protocol
from network import Network, InterfaceProblem
from binary import ports2byte, ports2list


def loglevel(x):
    try:
        return getattr(logging, x.upper())
    except AttributeError:
        raise argparse.ArgumentError("Select a proper loglevel")


def main():
    logger = logging.getLogger(__name__)
    parser = argparse.ArgumentParser()
    parser.add_argument("--switch-mac", "-s")
    parser.add_argument("--interface", "-i")
    parser.add_argument("--username", "-u")
    parser.add_argument("--password", "-p")
    parser.add_argument("--vlan", type=int)
    parser.add_argument("--vlan_name")
    parser.add_argument("--vlan_member")
    parser.add_argument("--vlan_tagged")
    parser.add_argument("--vlan_pvid")
    parser.add_argument("--delete", action="store_true")
    parser.add_argument("--loglevel", "-l", type=loglevel, default="INFO")
    parser.add_argument("action", default=None, nargs="?")
    args = parser.parse_args()

    logging.basicConfig(level=args.loglevel)
    actions = Protocol.tp_ids

    if args.action not in Protocol.tp_ids:
        print("Actions:\n")
        for action in Protocol.ids_tp.keys():
            if Protocol.ids_tp[action][2]:
                print(
                    "%s:%s %s [%s]"
                    % (
                        Protocol.ids_tp[action][1],
                        tabout(Protocol.ids_tp[action][1]),
                        Protocol.ids_tp[action][3],
                        Protocol.ids_tp[action][0],
                    )
                )
    else:
        net = None
        try:
            net = Network(args.interface, args.switch_mac)
        except InterfaceProblem as e:
            print("Error:", e)

        if net:
            net.login(args.username, args.password)
        else:
            exit(1)

        if args.action == "vlan" and args.vlan:
            if args.vlan_member or args.vlan_tagged or args.delete:
                if args.delete:
                    v = Protocol.set_vlan(int(args.vlan), 0, 0, "")
                else:
                    v = Protocol.set_vlan(
                        int(args.vlan),
                        ports2byte(args.vlan_member),
                        ports2byte(args.vlan_tagged),
                        args.vlan_name,
                    )
                l = [(actions["vlan"], v)]

            if args.vlan_pvid is not None:
                l = []
                for port in ports2list(args.vlan_pvid):
                    if port != 0:
                        l.append((actions["pvid"], Protocol.set_pvid(args.vlan, port)))
            header, payload = net.set(args.username, args.password, l)
        elif args.action in actions:
            header, payload = net.query(Protocol.GET, [(actions[args.action], b"")])
        print(*payload, sep="\n")


def tabout(cmd):
    # Format help output
    if len(cmd) < 8:
        return "\t\t"
    return "\t"


if __name__ == "__main__":
    main()



#!/usr/bin/env python

################################################################################################
# Listen to bidirectional network communication (to and from TP-Link switches / easy switch tool
# Analyze the network communications

import socket, time, random, argparse, logging
from protocol import Protocol
from network import Network


def show_packet(payload):
    print(payload)
    data = Protocol.decode(payload)
    print("TEST: " + data.hex(" "))
    try:
        print(Protocol.analyze(data)[1])
    except AssertionError:
        return
    except KeyError:
        return


def main():

    # Handle command-line parameters
    parser = argparse.ArgumentParser()
    parser.add_argument("--interface", "-i")
    args = parser.parse_args()

    # Instantiate network driver
    net = None
    try:
        net = Network(args.interface, "fe:ff:ff:ff:ff:ff")
    except InterfaceProblem as e:
        print("Error:", e)

    while True:
        data = net.receive_socket(net.rs)
        if data:
            show_packet(data)
        data = net.receive_socket(net.ss)
        if data:
            show_packet(data)

        time.sleep(0.1)


if __name__ == "__main__":
    main()



"""Provide protocol encode/decoding functions."""

import base64
import struct
from ipaddress import ip_address
from binary import byte2ports, mac_to_str


class Protocol:
    PACKET_END = b"\xff\xff\x00\x00"

    KEY_BASE64 = """
        v5vjymOiT2gxEr6kHky9gxc0VmrPfX6pxBysOryEoAMkeJCoDOd0LClhbNUqxiCU2mv3cMwOQkRb
        4M7rIYLLsgGGx075eweRSdDRZEpzSHYIFvOTQGAFVzxx6Zgf24+u6Jn1nv5GqktN19M7R4XWnZcG
        LlFeiKbSBCvxHd+wQz+6iYEo+P83Dz633mnsxX82s8LluSVa7bgZnK0au9wC4QDwMvvU/acRwc2x
        FbX2UuImZaO28lwUC18N5hB5fG3DdSdi71Q4i6EvyTOH+goTli1vGxiOUFVT6orYOV1Bmo16IoyA
        7lhZCZKrlTVmPXJF2a9n5CO0/MjApZ/d9G53MA==
    """

    KEY = base64.b64decode(KEY_BASE64)

    header = {
        "len": 32,
        "fmt": "!bb6s6shihhhhi",
        "blank": {
            "version": 1,
            "op_code": 0,
            "switch_mac": b"\x00\x00\x00\x00\x00\x00",
            "host_mac": b"\x00\x00\x00\x00\x00\x00",
            "sequence_id": 0,
            "error_code": 0,
            "check_length": 0,
            "fragment_offset": 0,
            "flag": 0,
            "token_id": 0,
            "checksum": 0,
        },
    }

    DISCOVERY = 0
    GET = 1
    SET = 2
    LOGIN = 3
    RETURN = 4
    READ5 = 5

    op_codes = {
        DISCOVERY: "DISCOVERY",
        GET: "GET",
        SET: "SET",
        LOGIN: "LOGIN",
        RETURN: "RETURN",
        READ5: "READ5",
    }

    sequences = {
        # name      ->switch    switch->
        "login/change": (LOGIN, RETURN),
        "query": (GET, SET),
        "discover": (DISCOVERY, SET),
    }

    ids_tp = {
        1: ("str", "type", 0),
        2: ("str", "hostname", 1, "Without arguments, shows switch details"),
        3: ("hex", "mac", 0),
        4: ("ip", "ip_addr", 0),
        5: ("ip", "ip_mask", 0),
        6: ("ip", "gateway", 0),
        7: ("str", "firmware", 0),
        8: ("str", "hardware", 0),
        9: ("bool", "dhcp", 0),
        10: ("dec", "num_ports", 0),
        12: ("bool", "led_status", 0),
        13: ("bool", "auto_save", 0),
        14: ("bool", "is_factory", 0),
        15: ("hex", "flash_type", 0),
        512: ("str", "username", 0),
        514: ("str", "password", 0),
        773: ("bool", "reboot", 1, "Reboot the device"),
        2304: ("action", "save", 1, "Saves the current configuration"),
        2305: ("action", "get_token_id", 1, "Unknown"),
        4352: ("bool", "igmp_snooping", 0),
        4096: ("hex", "ports", 0),
        4608: ("hex", "trunk", 0),
        8192: ("hex", "mtu_vlan", 0),
        8704: ("hex", "vlan_enabled", 0),
        8705: ("vlan", "vlan", 1, "Configure VLAN Membership"),
        8706: ("pvid", "pvid", 0),
        8707: ("str", "vlan_filler", 0),
        12288: ("bool", "qos1", 0),
        12289: ("hex", "qos2", 0),
        16640: ("hex", "mirror", 0),
        16384: ("stat", "stats", 0),
        17152: ("bool", "loop_prev", 0),
    }

    tp_ids = {v[1]: k for k, v in ids_tp.items()}

    def get_sequence_kind(sequence):
        for key, value in Protocol.sequences.items():
            if value == sequence:
                return key
        return "unknown"

    def get_id(name):
        return Protocol.tp_ids[name]

    def decode(data):
        data = bytearray(data)
        s = bytearray(Protocol.KEY)
        j = 0
        for k in range(len(data)):
            i = (k + 1) & 255
            j = (j + s[i]) & 255
            s[i], s[j] = s[j], s[i]
            data[k] = data[k] ^ s[(s[i] + s[j]) & 255]
        return bytes(data)

    encode = decode

    def split(data):
        if len(data) < Protocol.header["len"] + len(Protocol.PACKET_END):
            raise AssertionError("invalid data length")
        if not data.endswith(Protocol.PACKET_END):
            raise AssertionError("data without packet end")
        return (data[0 : Protocol.header["len"]], data[Protocol.header["len"] :])

    def interpret_header(header):
        names = Protocol.header["blank"].keys()
        vals = struct.unpack(Protocol.header["fmt"], header)
        return dict(zip(names, vals))

    def interpret_payload(payload):
        results = []
        while len(payload) > len(Protocol.PACKET_END):
            dtype, dlen = struct.unpack("!hh", payload[0:4])
            data = payload[4 : 4 + dlen]
            results.append(
                (
                    dtype,
                    Protocol.ids_tp[dtype][1],
                    Protocol.interpret_value(data, Protocol.ids_tp[dtype][0]),
                )
            )
            payload = payload[4 + dlen :]
        return results

    def analyze(data):
        header, payload = Protocol.split(data)
        return Protocol.interpret_header(header), Protocol.interpret_payload(payload)

    def assemble_packet(header, payload):
        payload_bytes = b""
        for dtype, value in payload:
            payload_bytes += struct.pack("!hh", dtype, len(value))
            payload_bytes += value
        header["check_length"] = (
            Protocol.header["len"] + len(payload_bytes) + len(Protocol.PACKET_END)
        )
        header = tuple(header[part] for part in Protocol.header["blank"].keys())
        header_bytes = struct.pack(Protocol.header["fmt"], *header)
        return header_bytes + payload_bytes + Protocol.PACKET_END

    def interpret_value(value, kind):
        if kind == "str":
            value = value.split(b"\x00", 1)[0].decode("ascii")
        elif kind == "ip":
            value = ip_address(value)
        elif kind == "hex":
            value = mac_to_str(value)
        elif kind == "action":
            value = "n/a"
        elif kind == "dec":
            value = int.from_bytes(value, "big")
        elif kind == "vlan":
            value = list(
                struct.unpack("!hii", value[:10]) + (value[10:-1].decode("ascii"),)
            )
            value[1] = byte2ports(value[1])
            value[2] = byte2ports(value[2])
        elif kind == "pvid":
            value = struct.unpack("!bh", value) if value else None
        elif kind == "stat":
            value = struct.unpack("!bbbIIII", value)
        elif kind == "bool":
            if len(value) == 0:
                pass
            elif len(value) == 1:
                value = value[0] > 0
            else:
                raise AssertionError("boolean should be one byte long")
        return value

    def set_vlan(vlan_num, member_mask, tagged_mask, vlan_name):
        value = (
            struct.pack("!hii", vlan_num, member_mask, tagged_mask)
            + vlan_name.encode("ascii")
            + b"\x00"
        )
        return value

    def set_pvid(vlan_num, port):
        value = struct.pack("!bh", port, vlan_num)
        return value



"""Provide a package for tplink-ess-lib."""
from __future__ import annotations

import logging

from .protocol import Protocol
from .network import Network, ConnectionProblem, InterfaceProblem

_LOGGER = logging.getLogger(__name__)

class MissingInterface(Exception):
    """Exception for missing interface."""


class tplink_ess:
    """Represent a tplink ess switch."""

    def __init__(
        self, mac: str = "", interface: str = "", user: str = "", pwd: str = ""
    ) -> None:
        """Connect or discover a tplink ess switch on the network."""
        self._user = user
        self._pwd = pwd
        self._mac = mac
        self._interface = interface
        self._data = {}

    async def discovery(self) -> dict:
        """Return result of auto discovery as dict."""
        if not self._interface:
            _LOGGER.error("Interface not setup.")
            raise MissingInterface
        net = Network(self._interface)
        net.send(Protocol.DISCOVERY, {})
        switches = {}
        i = 0
        while True:
            try:
                header, payload = net.receive()
                switches[i] = (header, payload)
            except ConnectionProblem:
                break
        return switches

    async def interfaces(self) -> list:
        """Return result of interface discovery."""
        interfaces = []
        interface = None
        net = Network(interface)
        interfaces = net.get_interface()
        return interfaces

    async def update_data(self) -> dict:
        """Refresh switch data."""
        try:
            net = Network(self._interface, self._mac)
        except InterfaceProblem as e:
            _LOGGER.error("Problems with network interface: %s", e)
            raise InterfaceProblem
        # Login to switch
        net.login(self._user, self._pwd)
        actions = Protocol.tp_ids

        for action in actions:
            header, payload = net.query(Protocol.GET, [(actions[action], b"")])
            self._data[action] = payload

        return self._data



#!/usr/bin/env python3
import logging


def loglevel(x):
    try:
        return getattr(logging, x.upper())
    except AttributeError:
        raise argparse.ArgumentError("Select a proper loglevel")


if __name__ == "__main__":
    pass
